from unittest.mock import patch

from django.test import override_settings
from eveuniverse.models import EveEntity

from app_utils.esi_testing import BravadoOperationStub
from app_utils.testing import NoSocketsTestCase

from ..tasks import run_manager_sync
from .factories import (
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
)
from .utils import (
    EsiCharacterContactsStub,
    EsiContact,
    EsiContactLabel,
    create_esi_contact,
)

MODELS_PATH = "standingssync.models"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODELS_PATH + ".STANDINGSSYNC_WAR_TARGETS_LABEL_NAME", "WAR TARGETS")
@patch(MODELS_PATH + ".esi")
class TestIntegration(NoSocketsTestCase):
    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    def test_should_sync_manager_and_character_no_wt(self, mock_esi):
        # given
        manager = SyncManagerFactory()
        sync_character = SyncedCharacterFactory(manager=manager)
        character = EveEntityCharacterFactory(
            id=sync_character.character.character_id,
            name=sync_character.character.character_name,
        )
        some_alliance_contact = EveEntityCharacterFactory()
        alliance_contacts = [
            create_esi_contact(character),
            create_esi_contact(some_alliance_contact),
        ]
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub(alliance_contacts)
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_esi_mock(mock_esi)
        # when
        run_manager_sync.delay(manager_pk=manager.pk)
        # then
        character_contacts = esi_character_contacts._contacts[
            sync_character.character.character_id
        ]
        self.assertEqual(character_contacts[some_alliance_contact.id].standing, 5)
        self.assertEqual(character_contacts[manager.alliance.alliance_id].standing, 10)
        self.assertNotIn(
            sync_character.character.character_id, character_contacts.keys()
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    def test_should_sync_manager_and_character_with_wt_as_defender(self, mock_esi):
        # given
        manager = SyncManagerFactory()
        alliance = EveEntity.objects.get(id=manager.alliance.alliance_id)
        war = EveWarFactory(defender=alliance)
        sync_character = SyncedCharacterFactory(manager=manager)
        character = EveEntityCharacterFactory(
            id=sync_character.character.character_id,
            name=sync_character.character.character_name,
        )
        some_alliance_contact = EveEntityCharacterFactory()
        alliance_contacts = [
            create_esi_contact(character),
            create_esi_contact(some_alliance_contact),
        ]
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub(alliance_contacts)
        )
        war_target_label = EsiContactLabel(1, "WAR TARGETS")
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_esi_mock(mock_esi)
        esi_character_contacts.setup_labels(character.id, [war_target_label])
        # when
        run_manager_sync.delay(manager_pk=manager.pk)
        # then
        result = set(
            esi_character_contacts.contacts(sync_character.character.character_id)
        )
        expected = {
            EsiContact.from_eve_entity(some_alliance_contact, standing=5),
            EsiContact(
                manager.alliance.alliance_id,
                EsiContact.ContactType.ALLIANCE,
                standing=10,
            ),
            EsiContact.from_eve_entity(
                war.aggressor, standing=-10, label_ids=[war_target_label.id]
            ),
        }
        self.assertSetEqual(result, expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    def test_should_sync_manager_and_character_with_wt_as_aggressor(self, mock_esi):
        # given
        manager = SyncManagerFactory()
        alliance = EveEntity.objects.get(id=manager.alliance.alliance_id)
        ally = EveEntityAllianceFactory()
        war = EveWarFactory(aggressor=alliance, allies=[ally])
        sync_character = SyncedCharacterFactory(manager=manager)
        character = EveEntityCharacterFactory(
            id=sync_character.character.character_id,
            name=sync_character.character.character_name,
        )
        some_alliance_contact = EveEntityCharacterFactory()
        some_character_contact = EveEntityCharacterFactory()
        alliance_contacts = [
            create_esi_contact(character),
            create_esi_contact(some_alliance_contact),
        ]
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub(alliance_contacts)
        )
        war_target_label = EsiContactLabel(1, "WAR TARGETS")
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_esi_mock(mock_esi)
        esi_character_contacts.setup_contacts(
            character.id,
            [
                EsiContact.from_eve_entity(ally, standing=5),
                EsiContact.from_eve_entity(some_character_contact, standing=10),
            ],
        )
        esi_character_contacts.setup_labels(character.id, [war_target_label])
        # when
        run_manager_sync.delay(manager_pk=manager.pk)
        # then
        result = set(
            esi_character_contacts.contacts(sync_character.character.character_id)
        )
        expected = {
            EsiContact.from_eve_entity(some_alliance_contact, standing=5),
            EsiContact(
                manager.alliance.alliance_id,
                EsiContact.ContactType.ALLIANCE,
                standing=10,
            ),
            EsiContact.from_eve_entity(
                war.defender, standing=-10, label_ids=[war_target_label.id]
            ),
            EsiContact.from_eve_entity(
                ally, standing=-10, label_ids=[war_target_label.id]
            ),
        }
        self.assertSetEqual(result, expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    def test_should_sync_manager_and_character_with_wt_as_aggressor_2(self, mock_esi):
        # given
        manager = SyncManagerFactory()
        alliance = EveEntity.objects.get(id=manager.alliance.alliance_id)
        ally = EveEntityAllianceFactory()
        war = EveWarFactory(aggressor=alliance, allies=[ally])
        sync_character = SyncedCharacterFactory(manager=manager)
        character = EveEntityCharacterFactory(
            id=sync_character.character.character_id,
            name=sync_character.character.character_name,
        )
        some_alliance_contact = EveEntityCharacterFactory()
        some_character_contact = EveEntityCharacterFactory()
        alliance_contacts = [
            create_esi_contact(character),
            create_esi_contact(some_alliance_contact),
        ]
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub(alliance_contacts)
        )
        war_target_label = EsiContactLabel(1, "WAR TARGETS")
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_esi_mock(mock_esi)
        esi_character_contacts.setup_contacts(
            character.id,
            [
                EsiContact.from_eve_entity(ally, standing=5),
                EsiContact.from_eve_entity(some_character_contact, standing=10),
            ],
        )
        esi_character_contacts.setup_labels(character.id, [war_target_label])
        # when
        run_manager_sync.delay(manager_pk=manager.pk)
        # then
        result = set(
            esi_character_contacts.contacts(sync_character.character.character_id)
        )
        expected = {
            EsiContact.from_eve_entity(
                war.defender, standing=-10, label_ids=[war_target_label.id]
            ),
            EsiContact.from_eve_entity(
                ally, standing=-10, label_ids=[war_target_label.id]
            ),
            EsiContact.from_eve_entity(some_character_contact, standing=10),
        }
        self.assertSetEqual(result, expected)



# Generated by Django 4.0.6 on 2022-07-29 17:08

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("standingssync", "0003_war_target_labels"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="evewar",
            name="aggressor",
        ),
        migrations.RemoveField(
            model_name="evewar",
            name="allies",
        ),
        migrations.RemoveField(
            model_name="evewar",
            name="defender",
        ),
        migrations.DeleteModel(
            name="EveContact",
        ),
        migrations.DeleteModel(
            name="EveEntity",
        ),
        migrations.DeleteModel(
            name="EveWar",
        ),
    ]



from app_utils.django import clean_setting

# minimum standing a character needs to have in order to get alliance contacts
# Any char with a standing smaller than this value will be rejected
STANDINGSSYNC_CHAR_MIN_STANDING = clean_setting(
    "STANDINGSSYNC_CHAR_MIN_STANDING", default_value=0.1, min_value=-10, max_value=10
)

# When enabled will automatically add or set war targets
# with standing = -10 to synced characters
STANDINGSSYNC_ADD_WAR_TARGETS = clean_setting("STANDINGSSYNC_ADD_WAR_TARGETS", False)

# Name of contacts label for war targets
STANDINGSSYNC_WAR_TARGETS_LABEL_NAME = clean_setting(
    "STANDINGSSYNC_WAR_TARGETS_LABEL_NAME", "WAR TARGETS"
)

# When enabled will replace contacts of synced characters with alliance contacts
STANDINGSSYNC_REPLACE_CONTACTS = clean_setting("STANDINGSSYNC_REPLACE_CONTACTS", True)

# Smallest war ID to fetch from ESI. All wars with smaller IDs are known to be already finished. This is an optimization to avoid having to fetch >700K wars from ESI.
STANDINGSSYNC_MINIMUM_UNFINISHED_WAR_ID = clean_setting(
    "STANDINGSSYNC_MINIMUM_UNFINISHED_WAR_ID", 719979
)

# IDs of unfinished wars, with IDs below the above minimum threshold
STANDINGSSYNC_SPECIAL_WAR_IDS = clean_setting(
    "STANDINGSSYNC_SPECIAL_WAR_IDS",
    [
        693125,
        716071,
        716072,
        716073,
        716864,
        717695,
        718307,
        718387,
        718575,
        718576,
        718619,
        718637,
        718638,
        718639,
        718640,
        718941,
        719186,
        719187,
        719188,
        719189,
        719226,
        719331,
        719336,
        719337,
        719423,
        719745,
        719751,
        719854,
        719890,
    ],
)



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from app_utils.esi_testing import BravadoOperationStub
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from ..managers import EveWarManager
from ..models import EveWar
from .factories import (
    EveContactFactory,
    EveEntityAllianceFactory,
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
)
from .utils import ALLIANCE_CONTACTS, LoadTestDataMixin

MANAGERS_PATH = "standingssync.managers"
MODELS_PATH = "standingssync.models"


class TestEveContactManager(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # 1 user with 1 alt character
        cls.user_1, _ = create_user_from_evecharacter(cls.character_1.character_id)
        cls.alt_ownership = CharacterOwnership.objects.create(
            character=cls.character_2, owner_hash="x2", user=cls.user_1
        )

        # sync manager with contacts
        cls.sync_manager = SyncManagerFactory(user=cls.user_1, version_hash="new")
        for contact in ALLIANCE_CONTACTS:
            EveContactFactory(
                manager=cls.sync_manager,
                eve_entity=EveEntity.objects.get(id=contact["contact_id"]),
                standing=contact["standing"],
            )

        # sync char
        cls.synced_character = SyncedCharacterFactory(
            character_ownership=cls.alt_ownership, manager=cls.sync_manager
        )

    def test_grouped_by_standing(self):
        c = {
            int(x.eve_entity_id): x
            for x in self.sync_manager.contacts.order_by("eve_entity_id")
        }
        expected = {
            -10.0: {c[1005], c[1012], c[3011], c[2011]},
            -5.0: {c[1013], c[3012], c[2012]},
            0.0: {c[1014], c[3013], c[2014]},
            5.0: {c[1015], c[3014], c[2013]},
            10.0: {c[1002], c[1004], c[1016], c[3015], c[2015]},
        }
        result = self.sync_manager.contacts.all().grouped_by_standing()
        self.maxDiff = None
        self.assertDictEqual(result, expected)


class TestEveWarManager(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # given
        cls.war_declared = now() - dt.timedelta(days=3)
        cls.war_started = now() - dt.timedelta(days=2)
        EveWarFactory(
            id=8,
            aggressor=EveEntity.objects.get(id=3011),
            defender=EveEntity.objects.get(id=3001),
            declared=cls.war_declared,
            started=cls.war_started,
            allies=[EveEntity.objects.get(id=3012)],
        )
        EveEntityAllianceFactory(id=3002)
        EveEntityAllianceFactory(id=3003)

    def test_should_return_defender_and_allies_for_aggressor(self):
        # when
        result = EveWar.objects.war_targets(3011)
        # then
        self.assertSetEqual({obj.id for obj in result}, {3001, 3012})

    def test_should_return_aggressor_for_defender(self):
        # when
        result = EveWar.objects.war_targets(3001)
        # then
        self.assertSetEqual({obj.id for obj in result}, {3011})

    def test_should_return_aggressor_for_ally(self):
        # when
        result = EveWar.objects.war_targets(3012)
        # then
        self.assertSetEqual({obj.id for obj in result}, {3011})

    def test_should_return_finished_wars(self):
        # given
        EveWarFactory(
            id=2,  # finished in the past
            aggressor=EveEntity.objects.get(id=3011),
            defender=EveEntity.objects.get(id=3001),
            declared=now() - dt.timedelta(days=5),
            started=now() - dt.timedelta(days=4),
            finished=now() - dt.timedelta(days=2),
        )
        EveWarFactory(
            id=3,  # about to finish
            aggressor=EveEntity.objects.get(id=3011),
            defender=EveEntity.objects.get(id=3001),
            declared=now() - dt.timedelta(days=5),
            started=now() - dt.timedelta(days=4),
            finished=now() + dt.timedelta(days=1),
        )
        EveWarFactory(
            id=4,  # not yet started
            aggressor=EveEntity.objects.get(id=3011),
            defender=EveEntity.objects.get(id=3001),
            declared=now() - dt.timedelta(days=1),
            started=now() + dt.timedelta(days=1),
        )
        # when
        result = EveWar.objects.finished_wars()
        # then
        self.assertSetEqual({obj.id for obj in result}, {2})

    @patch(MANAGERS_PATH + ".esi")
    def test_should_create_full_war_object_from_esi_1(self, mock_esi):
        # given
        declared = now() - dt.timedelta(days=5)
        started = now() - dt.timedelta(days=4)
        finished = now() + dt.timedelta(days=1)
        retracted = now()
        esi_data = {
            "aggressor": {
                "alliance_id": 3001,
                "isk_destroyed": 0,
                "ships_killed": 0,
            },
            "allies": [{"alliance_id": 3003}, {"corporation_id": 2003}],
            "declared": declared,
            "defender": {
                "alliance_id": 3002,
                "isk_destroyed": 0,
                "ships_killed": 0,
            },
            "finished": finished,
            "id": 1,
            "mutual": False,
            "open_for_allies": True,
            "retracted": retracted,
            "started": started,
        }
        mock_esi.client.Wars.get_wars_war_id.return_value = BravadoOperationStub(
            esi_data
        )
        # when
        EveWar.objects.update_or_create_from_esi(id=1)
        # then
        self.assertTrue(EveWar.objects.filter(id=1).exists())
        war = EveWar.objects.get(id=1)
        self.assertEqual(war.aggressor.id, 3001)
        self.assertEqual(set(war.allies.values_list("id", flat=True)), {2003, 3003})
        self.assertEqual(war.declared, declared)
        self.assertEqual(war.defender.id, 3002)
        self.assertEqual(war.finished, finished)
        self.assertFalse(war.is_mutual)
        self.assertTrue(war.is_open_for_allies)
        self.assertEqual(war.retracted, retracted)
        self.assertEqual(war.started, started)

    @patch(MANAGERS_PATH + ".esi")
    def test_should_create_full_war_object_from_esi_2(self, mock_esi):
        # given
        declared = now() - dt.timedelta(days=5)
        started = now() - dt.timedelta(days=4)
        esi_data = {
            "aggressor": {
                "alliance_id": 3001,
                "isk_destroyed": 0,
                "ships_killed": 0,
            },
            "allies": None,
            "declared": declared,
            "defender": {
                "alliance_id": 3002,
                "isk_destroyed": 0,
                "ships_killed": 0,
            },
            "finished": None,
            "id": 1,
            "mutual": False,
            "open_for_allies": True,
            "retracted": None,
            "started": started,
        }
        mock_esi.client.Wars.get_wars_war_id.return_value = BravadoOperationStub(
            esi_data
        )
        # when
        EveWar.objects.update_or_create_from_esi(id=1)
        # then
        self.assertTrue(EveWar.objects.filter(id=1).exists())
        war = EveWar.objects.get(id=1)
        self.assertEqual(war.aggressor.id, 3001)
        self.assertEqual(war.allies.count(), 0)
        self.assertEqual(war.declared, declared)
        self.assertEqual(war.defender.id, 3002)
        self.assertIsNone(war.finished)
        self.assertFalse(war.is_mutual)
        self.assertTrue(war.is_open_for_allies)
        self.assertIsNone(war.retracted)
        self.assertEqual(war.started, started)

    # @patch(MANAGERS_PATH + ".esi")
    # def test_should_not_create_object_from_esi_for_finished_war(self, mock_esi):
    #     # given
    #     declared = now() - dt.timedelta(days=5)
    #     started = now() - dt.timedelta(days=4)
    #     finished = now() - dt.timedelta(days=1)
    #     esi_data = {
    #         "aggressor": {
    #             "alliance_id": 3001,
    #             "isk_destroyed": 0,
    #             "ships_killed": 0,
    #         },
    #         "allies": [{"alliance_id": 3003}, {"corporation_id": 2003}],
    #         "declared": declared,
    #         "defender": {
    #             "alliance_id": 3002,
    #             "isk_destroyed": 0,
    #             "ships_killed": 0,
    #         },
    #         "finished": finished,
    #         "id": 1,
    #         "mutual": False,
    #         "open_for_allies": True,
    #         "retracted": None,
    #         "started": started,
    #     }
    #     mock_esi.client.Wars.get_wars_war_id.return_value = BravadoOperationStub(
    #         esi_data
    #     )
    #     # when
    #     EveWar.objects.update_or_create_from_esi(id=1)
    #     # then
    #     self.assertFalse(EveWar.objects.filter(id=1).exists())

    @patch(MANAGERS_PATH + ".esi")
    def test_should_update_existing_war_from_esi(self, mock_esi):
        # given
        finished = now() + dt.timedelta(days=1)
        retracted = now()
        esi_data = {
            "aggressor": {
                "alliance_id": 3011,
                "isk_destroyed": 0,
                "ships_killed": 0,
            },
            "allies": [{"alliance_id": 3003}, {"corporation_id": 2003}],
            "declared": self.war_declared,
            "defender": {
                "alliance_id": 3001,
                "isk_destroyed": 0,
                "ships_killed": 0,
            },
            "finished": finished,
            "id": 8,
            "mutual": True,
            "open_for_allies": True,
            "retracted": retracted,
            "started": self.war_started,
        }
        mock_esi.client.Wars.get_wars_war_id.return_value = BravadoOperationStub(
            esi_data
        )
        # when
        EveWar.objects.update_or_create_from_esi(id=8)
        # then
        self.assertTrue(EveWar.objects.filter(id=8).exists())
        war = EveWar.objects.get(id=8)
        self.assertEqual(war.aggressor.id, 3011)
        self.assertEqual(set(war.allies.values_list("id", flat=True)), {2003, 3003})
        self.assertEqual(war.declared, self.war_declared)
        self.assertEqual(war.defender.id, 3001)
        self.assertEqual(war.finished, finished)
        self.assertTrue(war.is_mutual)
        self.assertTrue(war.is_open_for_allies)
        self.assertEqual(war.retracted, retracted)
        self.assertEqual(war.started, self.war_started)


class TestEveWarManager2(NoSocketsTestCase):
    @patch(MANAGERS_PATH + ".STANDINGSSYNC_SPECIAL_WAR_IDS", [3, 4])
    @patch(MODELS_PATH + ".EveWar.objects.fetch_war_ids_from_esi")
    def test_should_return_relevant_war_ids(self, mock_fetch_war_ids_from_esi):
        # given
        mock_fetch_war_ids_from_esi.return_value = {1, 2, 42}
        EveWarFactory(id=42, finished=now() - dt.timedelta(days=1))
        # when
        result = EveWar.objects.calc_relevant_war_ids()
        # then
        self.assertSetEqual(result, {1, 2, 3, 4})

    @patch(MANAGERS_PATH + ".STANDINGSSYNC_MINIMUM_UNFINISHED_WAR_ID", 4)
    @patch(MANAGERS_PATH + ".esi")
    def test_should_fetch_war_ids_with_paging(self, mock_esi):
        def esi_get_wars(max_war_id=None):
            if max_war_id:
                war_ids = [war_id for war_id in esi_war_ids if war_id < max_war_id]
            else:
                war_ids = esi_war_ids
            return BravadoOperationStub(sorted(war_ids, reverse=True)[:page_size])

        # given
        esi_war_ids = [1, 2, 3, 4, 5, 6, 7, 8]
        page_size = 3
        mock_esi.client.Wars.get_wars.side_effect = esi_get_wars
        # when
        result = EveWarManager.fetch_war_ids_from_esi(max_items=3)
        # then
        self.assertSetEqual(result, {4, 5, 6, 7, 8})


class TestEveWarManagerActiveWars(NoSocketsTestCase):
    def test_should_return_started_war_as_defender(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_started_war_as_attacker(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            aggressor=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_started_war_as_ally(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            allies=[EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id)],
            declared=now() - dt.timedelta(days=2),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_return_war_about_to_finish(self):
        # given
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
            finished=now() + dt.timedelta(days=1),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 1)
        self.assertEqual(result.first(), war)

    def test_should_not_return_finished_war(self):
        # given
        sync_manager = SyncManagerFactory()
        EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=2),
            finished=now() - dt.timedelta(days=1),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 0)

    def test_should_not_return_war_not_yet_started(self):
        # given
        sync_manager = SyncManagerFactory()
        EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            declared=now() - dt.timedelta(days=1),
            started=now() + dt.timedelta(hours=4),
        )
        # when
        result = EveWar.objects.active_wars()
        # then
        self.assertEqual(result.count(), 0)



from typing import List, Set

from django.db import models, transaction
from django.db.models import Exists, OuterRef
from django.utils.timezone import now

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    STANDINGSSYNC_MINIMUM_UNFINISHED_WAR_ID,
    STANDINGSSYNC_SPECIAL_WAR_IDS,
)
from .providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveContactQuerySet(models.QuerySet):
    def grouped_by_standing(self) -> dict:
        """Return alliance contacts grouped by their standing as dict."""
        contacts_by_standing = dict()
        for contact in self.all():
            if contact.standing not in contacts_by_standing:
                contacts_by_standing[contact.standing] = set()
            contacts_by_standing[contact.standing].add(contact)
        return contacts_by_standing


class EveContactManagerBase(models.Manager):
    pass


EveContactManager = EveContactManagerBase.from_queryset(EveContactQuerySet)


class EveWarQuerySet(models.QuerySet):
    def annotate_active_wars(self) -> models.QuerySet:
        from .models import EveWar

        return self.annotate(
            active=Exists(EveWar.objects.active_wars().filter(pk=OuterRef("pk")))
        )

    def active_wars(self) -> models.QuerySet:
        return self.filter(started__lt=now(), finished__gt=now()) | self.filter(
            started__lt=now(), finished__isnull=True
        )

    def finished_wars(self) -> models.QuerySet:
        return self.filter(finished__lte=now())


class EveWarManagerBase(models.Manager):
    def war_targets(self, alliance_id: int) -> List[models.Model]:
        """Return list of current war targets for given alliance as EveEntity objects
        or an empty list if there are None.
        """
        war_targets = list()
        active_wars = self.active_wars()
        # case 1 alliance is aggressor
        for war in active_wars:
            if war.aggressor_id == alliance_id:
                war_targets.append(war.defender)
                if war.allies:
                    war_targets += list(war.allies.all())

        # case 2 alliance is defender
        for war in active_wars:
            if war.defender_id == alliance_id:
                war_targets.append(war.aggressor)

        # case 3 alliance is ally
        for war in active_wars:
            if war.allies.filter(id=alliance_id).exists():
                war_targets.append(war.aggressor)

        return war_targets

    def update_or_create_from_esi(self, id: int):
        """Updates existing or creates new objects from ESI with given ID."""
        from .models import EveEntity

        logger.info("Retrieving war details for ID %s", id)
        war_info = esi.client.Wars.get_wars_war_id(war_id=id).results(ignore_cache=True)
        aggressor, _ = EveEntity.objects.get_or_create(
            id=self._extract_id_from_war_participant(war_info["aggressor"])
        )
        defender, _ = EveEntity.objects.get_or_create(
            id=self._extract_id_from_war_participant(war_info["defender"])
        )
        with transaction.atomic():
            war, _ = self.update_or_create(
                id=id,
                defaults={
                    "aggressor": aggressor,
                    "declared": war_info["declared"],
                    "defender": defender,
                    "is_mutual": war_info["mutual"],
                    "is_open_for_allies": war_info["open_for_allies"],
                    "retracted": war_info.get("retracted"),
                    "started": war_info.get("started"),
                    "finished": war_info.get("finished"),
                },
            )
            war.allies.clear()
            if war_info.get("allies"):
                for ally_info in war_info.get("allies"):
                    eve_entity, _ = EveEntity.objects.get_or_create(
                        id=self._extract_id_from_war_participant(ally_info)
                    )
                    war.allies.add(eve_entity)

    @staticmethod
    def _extract_id_from_war_participant(participant: dict) -> int:
        alliance_id = participant.get("alliance_id")
        corporation_id = participant.get("corporation_id")
        if not alliance_id and not corporation_id:
            raise ValueError(f"Invalid participant: {participant}")
        return alliance_id or corporation_id

    def calc_relevant_war_ids(self) -> Set[int]:
        """Determine IDs from unfinished and new wars."""
        logger.info("Fetching wars from ESI")
        war_ids = self.fetch_war_ids_from_esi()
        war_ids = war_ids.union(set(STANDINGSSYNC_SPECIAL_WAR_IDS))
        finished_war_ids = set(self.finished_wars().values_list("id", flat=True))
        war_ids = set(war_ids)
        return war_ids.difference(finished_war_ids)

    @staticmethod
    def fetch_war_ids_from_esi(max_items: int = 2000) -> Set[int]:
        """Fetch IDs for new and unfinished wars from ESI.

        Will ignore older wars which are known to be already finished.
        """
        logger.info("Fetching war IDs from ESI")
        war_ids = []
        war_ids_page = esi.client.Wars.get_wars().results(ignore_cache=True)
        while True:
            war_ids += war_ids_page
            if (
                len(war_ids_page) < max_items
                or min(war_ids_page) < STANDINGSSYNC_MINIMUM_UNFINISHED_WAR_ID
            ):
                break
            max_war_id = min(war_ids)
            war_ids_page = esi.client.Wars.get_wars(max_war_id=max_war_id).results(
                ignore_cache=True
            )
        return set(
            [
                war_id
                for war_id in war_ids
                if war_id >= STANDINGSSYNC_MINIMUM_UNFINISHED_WAR_ID
            ]
        )


EveWarManager = EveWarManagerBase.from_queryset(EveWarQuerySet)



import hashlib
import json
from typing import Optional

from django.db import models, transaction
from django.utils.timezone import now
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    STANDINGSSYNC_ADD_WAR_TARGETS,
    STANDINGSSYNC_CHAR_MIN_STANDING,
    STANDINGSSYNC_REPLACE_CONTACTS,
    STANDINGSSYNC_WAR_TARGETS_LABEL_NAME,
)
from .managers import EveContactManager, EveWarManager
from .providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class _SyncBaseModel(models.Model):
    """Base for sync models"""

    version_hash = models.CharField(max_length=32, default="")
    last_sync = models.DateTimeField(null=True, default=None)

    class Meta:
        abstract = True

    @property
    def is_sync_ok(self) -> bool:
        return self.last_error == self.Error.NONE

    def set_sync_status(self, status: int) -> None:
        """sets the sync status with the current date and time"""
        self.last_error = status
        self.last_sync = now()
        self.save()


class SyncManager(_SyncBaseModel):
    """An object for managing syncing of contacts for an alliance"""

    class Error(models.IntegerChoices):
        NONE = 0, "No error"
        TOKEN_INVALID = 1, "Invalid token"
        TOKEN_EXPIRED = 2, "Expired token"
        INSUFFICIENT_PERMISSIONS = 3, "Insufficient permissions"
        NO_CHARACTER = 4, "No character set for fetching alliance contacts"
        ESI_UNAVAILABLE = 5, "ESI API is currently unavailable"
        UNKNOWN = 99, "Unknown error"

    alliance = models.OneToOneField(
        EveAllianceInfo, on_delete=models.CASCADE, primary_key=True, related_name="+"
    )
    # alliance contacts are fetched from this character
    character_ownership = models.OneToOneField(
        CharacterOwnership, on_delete=models.SET_NULL, null=True, default=None
    )
    last_error = models.IntegerField(choices=Error.choices, default=Error.NONE)

    def __str__(self):
        if self.character_ownership is not None:
            character_name = self.character_ownership.character.character_name
        else:
            character_name = "None"
        return "{} ({})".format(self.alliance.alliance_name, character_name)

    def get_effective_standing(self, character: EveCharacter) -> float:
        """return the effective standing with this alliance"""

        contact_found = None
        try:
            contact_found = self.contacts.get(eve_entity_id=character.character_id)
        except EveContact.DoesNotExist:
            try:
                contact_found = self.contacts.get(
                    eve_entity_id=character.corporation_id
                )
            except EveContact.DoesNotExist:
                if character.alliance_id:
                    try:
                        contact_found = self.contacts.get(
                            eve_entity_id=character.alliance_id
                        )
                    except EveContact.DoesNotExist:
                        pass

        return contact_found.standing if contact_found is not None else 0.0

    def update_from_esi(self, force_sync: bool = False) -> Optional[str]:
        """Update this sync manager from ESi

        Args:
        - force_sync: will ignore version_hash if set to true

        Returns:
        - newest version hash on success or None on error
        """
        if self.character_ownership is None:
            logger.error("%s: No character configured to sync the alliance", self)
            self.set_sync_status(self.Error.NO_CHARACTER)
            return None

        # abort if character does not have sufficient permissions
        if not self.character_ownership.user.has_perm("standingssync.add_syncmanager"):
            logger.error(
                "%s: Character does not have sufficient permission "
                "to sync the alliance",
                self,
            )
            self.set_sync_status(self.Error.INSUFFICIENT_PERMISSIONS)
            return None

        try:
            # get token
            token = (
                Token.objects.filter(
                    user=self.character_ownership.user,
                    character_id=self.character_ownership.character.character_id,
                )
                .require_scopes(self.get_esi_scopes())
                .require_valid()
                .first()
            )

        except TokenInvalidError:
            logger.error("%s: Invalid token for fetching alliance contacts", self)
            self.set_sync_status(self.Error.TOKEN_INVALID)
            return None

        except TokenExpiredError:
            self.set_sync_status(self.Error.TOKEN_EXPIRED)
            return None

        else:
            if not token:
                self.set_sync_status(self.Error.TOKEN_INVALID)
                return None

        new_version_hash = self._perform_update_from_esi(token, force_sync)
        self.set_sync_status(self.Error.NONE)
        return new_version_hash

    def _perform_update_from_esi(self, token, force_sync) -> str:
        # get alliance contacts
        alliance_id = self.character_ownership.character.alliance_id
        contacts_raw = esi.client.Contacts.get_alliances_alliance_id_contacts(
            token=token.valid_access_token(), alliance_id=alliance_id
        ).results()
        contacts = {int(row["contact_id"]): row for row in contacts_raw}

        if STANDINGSSYNC_ADD_WAR_TARGETS:
            war_targets = EveWar.objects.war_targets(alliance_id)
            for war_target in war_targets:
                contacts[war_target.id] = self._to_esi_dict(war_target, -10.0)
            war_target_ids = {war_target.id for war_target in war_targets}
        else:
            war_target_ids = set()

        # determine if contacts have changed by comparing their hashes
        new_version_hash = self._calculate_version_hash(contacts)
        if force_sync or new_version_hash != self.version_hash:
            logger.info(
                "%s: Storing alliance update with %d contacts", self, len(contacts)
            )

            # add the sync alliance with max standing to contacts
            contacts[alliance_id] = {
                "contact_id": alliance_id,
                "contact_type": "alliance",
                "standing": 10,
            }
            with transaction.atomic():
                self.version_hash = new_version_hash
                self.save()
                self.contacts.all().delete()
                contacts = [
                    EveContact(
                        manager=self,
                        eve_entity=EveEntity.objects.get_or_create(id=contact_id)[0],
                        standing=contact["standing"],
                        is_war_target=contact_id in war_target_ids,
                    )
                    for contact_id, contact in contacts.items()
                ]
                EveContact.objects.bulk_create(contacts, batch_size=500)
        else:
            logger.info("%s: Alliance contacts are unchanged.", self)
        return new_version_hash

    @staticmethod
    def _to_esi_dict(eve_entity: EveEntity, standing: float) -> dict:
        """Convert EveEntity to ESI contact dict."""
        return {
            "contact_id": eve_entity.id,
            "contact_type": eve_entity.category,
            "standing": standing,
        }

    @staticmethod
    def _calculate_version_hash(contacts: dict) -> str:
        """Calculate hash for contacts."""
        return hashlib.md5(json.dumps(contacts).encode("utf-8")).hexdigest()

    @classmethod
    def get_esi_scopes(cls) -> list:
        return ["esi-alliances.read_contacts.v1"]


class SyncedCharacter(_SyncBaseModel):
    """A character that has his personal contacts synced with an alliance"""

    class Error(models.IntegerChoices):
        NONE = 0, "No error"
        TOKEN_INVALID = 1, "Invalid token"
        TOKEN_EXPIRED = 2, "Expired token"
        INSUFFICIENT_PERMISSIONS = 3, "Insufficient permissions"
        ESI_UNAVAILABLE = 5, "ESI API is currently unavailable"
        UNKNOWN = 99, "Unknown error"

    character_ownership = models.OneToOneField(
        CharacterOwnership, on_delete=models.CASCADE, primary_key=True
    )
    manager = models.ForeignKey(
        SyncManager, on_delete=models.CASCADE, related_name="synced_characters"
    )
    last_error = models.IntegerField(choices=Error.choices, default=Error.NONE)
    has_war_targets_label = models.BooleanField(default=None, null=True)

    def __str__(self):
        return self.character_ownership.character.character_name

    @property
    def character(self) -> int:
        return self.character_ownership.character

    def get_status_message(self):
        if self.last_error != self.Error.NONE:
            return self.get_last_error_display()
        elif self.last_sync is not None:
            return "OK"
        return "Not synced yet"

    def update(self, force_sync: bool = False) -> bool:
        """updates in-game contacts for given character

        Will delete the sync character if necessary,
        e.g. if token is no longer valid or character is no longer blue

        Args:
        - force_sync: will ignore version_hash if set to true

        Returns:
        - False if the sync character was deleted, True otherwise
        """
        # abort if owner does not have sufficient permissions
        logger.info("%s: Updating contacts", self)
        if not self.character_ownership.user.has_perm(
            "standingssync.add_syncedcharacter"
        ):
            logger.info(
                "%s: sync deactivated due to insufficient user permissions", self
            )
            self._deactivate_sync("you no longer have permission for this service")
            return False

        # check if an update is needed
        if not force_sync and self.manager.version_hash == self.version_hash:
            logger.info(
                "%s: contacts of this char are up-to-date, no sync required", self
            )
            return True

        token = self._fetch_token()
        if not token:
            return False

        if not self.manager.contacts.exists():
            logger.info("%s: No contacts to sync", self)
            return True

        character_eff_standing = self.manager.get_effective_standing(
            self.character_ownership.character
        )
        if character_eff_standing < STANDINGSSYNC_CHAR_MIN_STANDING:
            logger.info(
                "%s: sync deactivated because character is no longer considered blue. "
                f"It's standing is: {character_eff_standing}, "
                f"while STANDINGSSYNC_CHAR_MIN_STANDING is: {STANDINGSSYNC_CHAR_MIN_STANDING} ",
                self,
            )
            self._deactivate_sync(
                "your character is no longer blue with the alliance. "
                f"The standing value is: {character_eff_standing:.1f} "
            )
            return False

        character_id = self.character_ownership.character.character_id
        logger.info("%s: Fetching current contacts", self)
        character_contacts_raw = (
            esi.client.Contacts.get_characters_character_id_contacts(
                token=token.valid_access_token(), character_id=character_id
            ).results()
        )
        character_contacts = {
            contact["contact_id"]: contact for contact in character_contacts_raw
        }
        logger.info("%s: Fetching current labels", self)
        labels_raw = esi.client.Contacts.get_characters_character_id_contacts_labels(
            character_id=character_id, token=token.valid_access_token()
        ).results()
        war_target_id = self._determine_war_target_id(labels_raw)
        if war_target_id:
            logger.debug("%s: Has war target label", self)
            self.has_war_targets_label = True
            self.save()
            ids_to_delete = [
                contact_id
                for contact_id, contact in character_contacts.items()
                if contact["label_ids"] and war_target_id in contact["label_ids"]
            ]
            if ids_to_delete:
                logger.info("%s: Removing old war target contacts", self)
                self._esi_delete_contacts(
                    character_id=character_id, token=token, contact_ids=ids_to_delete
                )
                for contact_id in ids_to_delete:
                    del character_contacts[contact_id]
        else:
            logger.debug("%s: Does not have war target label", self)
            self.has_war_targets_label = False
            self.save()

        if STANDINGSSYNC_REPLACE_CONTACTS:
            logger.info("%s: Deleting current contacts", self)
            self._esi_delete_contacts(
                character_id=character_id,
                token=token,
                contact_ids=list(character_contacts.keys()),
            )
            if STANDINGSSYNC_ADD_WAR_TARGETS and war_target_id:
                logger.info("%s: Writing alliance contacts and war targets", self)
                self._esi_update(
                    character_id=character_id,
                    token=token,
                    contacts_by_standing=self.manager.contacts.exclude(
                        eve_entity_id=character_id
                    )
                    .filter(is_war_target=False)
                    .grouped_by_standing(),
                    esi_method=esi.client.Contacts.post_characters_character_id_contacts,
                )
                self._esi_update(
                    character_id=character_id,
                    token=token,
                    contacts_by_standing=self.manager.contacts.filter(
                        is_war_target=True
                    ).grouped_by_standing(),
                    esi_method=esi.client.Contacts.post_characters_character_id_contacts,
                    label_ids=[war_target_id],
                )
            else:
                logger.info("%s: Writing alliance contacts", self)
                self._esi_update(
                    character_id=character_id,
                    token=token,
                    contacts_by_standing=self.manager.contacts.exclude(
                        eve_entity_id=character_id
                    ).grouped_by_standing(),
                    esi_method=esi.client.Contacts.post_characters_character_id_contacts,
                )
        else:
            contacts = self.manager.contacts.filter(is_war_target=True)
            if contacts.exists():
                logger.info("%s: Update existing contacts to war target", self)
                contacts_by_standing = contacts.filter(
                    eve_entity_id__in=character_contacts.keys()
                ).grouped_by_standing()
                self._esi_update(
                    character_id=character_id,
                    token=token,
                    contacts_by_standing=contacts_by_standing,
                    esi_method=esi.client.Contacts.put_characters_character_id_contacts,
                    label_ids=[war_target_id] if war_target_id else None,
                )
                logger.info("%s: Add new war target contacts", self)
                contacts_by_standing = contacts.exclude(
                    eve_entity_id__in=character_contacts.keys()
                ).grouped_by_standing()
                self._esi_update(
                    character_id=character_id,
                    token=token,
                    contacts_by_standing=contacts_by_standing,
                    esi_method=esi.client.Contacts.post_characters_character_id_contacts,
                    label_ids=[war_target_id] if war_target_id else None,
                )

        # store updated version hash with character
        self.version_hash = self.manager.version_hash
        self.save()
        self.set_sync_status(self.Error.NONE)
        return True

    def _determine_war_target_id(self, labels_raw: list) -> Optional[int]:
        for row in labels_raw:
            if (
                row.get("label_name").lower()
                == STANDINGSSYNC_WAR_TARGETS_LABEL_NAME.lower()
            ):
                war_target_id = row.get("label_id")
                break
        else:
            war_target_id = None
        return war_target_id

    @staticmethod
    def _esi_delete_contacts(character_id: int, token: Token, contact_ids: list):
        max_items = 20
        contact_ids_chunks = chunks(contact_ids, max_items)
        for contact_ids_chunk in contact_ids_chunks:
            esi.client.Contacts.delete_characters_character_id_contacts(
                token=token.valid_access_token(),
                character_id=character_id,
                contact_ids=contact_ids_chunk,
            ).results()

    @staticmethod
    def _esi_update(
        character_id: int,
        token: Token,
        contacts_by_standing: dict,
        esi_method,
        label_ids: list = None,
        max_items: int = 100,
    ):
        for standing in contacts_by_standing:
            contact_ids_chunks = chunks(
                [contact.eve_entity_id for contact in contacts_by_standing[standing]],
                max_items,
            )
            for contact_ids_chunk in contact_ids_chunks:
                esi_method(
                    token=token.valid_access_token(),
                    character_id=character_id,
                    contact_ids=contact_ids_chunk,
                    standing=standing,
                    label_ids=label_ids if label_ids else [],
                ).results()

    def _fetch_token(self) -> Optional[Token]:
        try:
            token = (
                Token.objects.filter(
                    user=self.character_ownership.user,
                    character_id=self.character_ownership.character.character_id,
                )
                .require_scopes(self.get_esi_scopes())
                .require_valid()
                .first()
            )
        except TokenInvalidError:
            logger.info("%s: sync deactivated due to invalid token", self)
            self._deactivate_sync("your token is no longer valid")
            return None

        except TokenExpiredError:
            logger.info("%s: sync deactivated due to expired token", self)
            self._deactivate_sync("your token has expired")
            return None

        if token is None:
            logger.info("%s: can not find suitable token for synced char", self)
            self._deactivate_sync("you do not have a token anymore")
            return None

        return token

    def _deactivate_sync(self, message):
        message = (
            "Standings Sync has been deactivated for your "
            f"character {self}, because {message}.\n"
            "Feel free to activate sync for your character again, "
            "once the issue has been resolved."
        )
        notify(
            self.character_ownership.user,
            f"Standings Sync deactivated for {self}",
            message,
        )
        self.delete()

    @staticmethod
    def get_esi_scopes() -> list:
        return ["esi-characters.read_contacts.v1", "esi-characters.write_contacts.v1"]


class EveContact(models.Model):
    """An Eve Online contact"""

    manager = models.ForeignKey(
        SyncManager, on_delete=models.CASCADE, related_name="contacts"
    )
    eve_entity = models.ForeignKey(
        EveEntity, on_delete=models.CASCADE, related_name="+"
    )
    standing = models.FloatField()
    is_war_target = models.BooleanField()

    objects = EveContactManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["manager", "eve_entity"], name="fk_eve_contact"
            )
        ]

    def __str__(self):
        return f"{self.eve_entity}"


class EveWar(models.Model):
    """An EveOnline war"""

    id = models.PositiveIntegerField(primary_key=True)
    aggressor = models.ForeignKey(EveEntity, on_delete=models.CASCADE, related_name="+")
    allies = models.ManyToManyField(EveEntity, related_name="+")
    declared = models.DateTimeField()
    defender = models.ForeignKey(EveEntity, on_delete=models.CASCADE, related_name="+")
    finished = models.DateTimeField(null=True, default=None, db_index=True)
    is_mutual = models.BooleanField()
    is_open_for_allies = models.BooleanField()
    retracted = models.DateTimeField(null=True, default=None)
    started = models.DateTimeField(null=True, default=None, db_index=True)

    objects = EveWarManager()

    def __str__(self) -> str:
        return f"{self.aggressor} vs. {self.defender}"



from celery import shared_task

from eveuniverse.core.esitools import is_esi_online
from eveuniverse.models import EveEntity
from eveuniverse.tasks import update_unresolved_eve_entities

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .models import EveWar, SyncedCharacter, SyncManager

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


DEFAULT_TASK_PRIORITY = 6


@shared_task
def run_regular_sync():
    """update all wars, managers and related characters if needed"""
    if not is_esi_online():
        logger.warning("ESI is not online. aborting")
        return

    update_all_wars.apply_async(priority=DEFAULT_TASK_PRIORITY)
    for sync_manager_pk in SyncManager.objects.values_list("pk", flat=True):
        run_manager_sync.apply_async(
            args=[sync_manager_pk], priority=DEFAULT_TASK_PRIORITY
        )


@shared_task
def run_manager_sync(manager_pk: int, force_sync: bool = False) -> bool:
    """updates contacts for given manager and related characters

    Args:
    - manage_pk: primary key of sync manager to run sync for
    - force_sync: will ignore version_hash if set to true

    Returns:
    - True on success or False on error
    """

    sync_manager = SyncManager.objects.get(pk=manager_pk)
    try:
        new_version_hash = sync_manager.update_from_esi(force_sync)
    except Exception:
        logger.debug("Unexpected exception occurred", exc_info=True)
        sync_manager.set_sync_status(sync_manager.Error.UNKNOWN)
        return False
    else:
        EveEntity.objects.bulk_update_new_esi()

    if not new_version_hash:
        return False

    if force_sync:
        alts_need_syncing = sync_manager.synced_characters.values_list("pk", flat=True)
    else:
        alts_need_syncing = sync_manager.synced_characters.exclude(
            version_hash=new_version_hash
        ).values_list("pk", flat=True)
    for character_pk in alts_need_syncing:
        run_character_sync.apply_async(
            kwargs={"sync_char_pk": character_pk, "force_sync": force_sync},
            priority=DEFAULT_TASK_PRIORITY,
        )

    return True


@shared_task
def run_character_sync(sync_char_pk: int, force_sync: bool = False) -> bool:
    """updates in-game contacts for given character

    Args:
    - sync_char_pk: primary key of sync character to run sync for
    - force_sync: will ignore version_hash if set to true

    Returns:
    - False if sync failed and the sync character was deleted, True otherwise
    """

    synced_character = SyncedCharacter.objects.get(pk=sync_char_pk)
    try:
        return synced_character.update(force_sync)
    except Exception as ex:
        logger.error("An unexpected error ocurred: %s", ex, exc_info=True)
        synced_character.set_sync_status(SyncedCharacter.Error.UNKNOWN)
        raise ex


@shared_task
def update_all_wars():
    relevant_war_ids = EveWar.objects.calc_relevant_war_ids()
    logger.info("Fetching details for %s wars from ESI", len(relevant_war_ids))
    for war_id in relevant_war_ids:
        update_war.apply_async(args=[war_id], priority=DEFAULT_TASK_PRIORITY)
    update_unresolved_eve_entities.apply_async(priority=DEFAULT_TASK_PRIORITY)


@shared_task
def update_war(war_id: int):
    EveWar.objects.update_or_create_from_esi(war_id)



from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render
from django.utils.html import format_html
from esi.decorators import token_required

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag
from app_utils.messages import messages_plus
from app_utils.views import link_html

from . import __title__, tasks
from .app_settings import (
    STANDINGSSYNC_ADD_WAR_TARGETS,
    STANDINGSSYNC_CHAR_MIN_STANDING,
    STANDINGSSYNC_REPLACE_CONTACTS,
    STANDINGSSYNC_WAR_TARGETS_LABEL_NAME,
)
from .models import SyncedCharacter, SyncManager

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def create_img_html(src: str, classes: list = None, size: int = None) -> str:
    classes_str = format_html('class="{}"', (" ".join(classes)) if classes else "")
    size_html = format_html('width="{}" height="{}"', size, size) if size else ""
    return format_html('<img {} {} src="{}">', classes_str, size_html, src)


def create_icon_plus_name_html(
    icon_url,
    name,
    size: int = 32,
    avatar: bool = False,
    url: str = None,
    text: str = None,
) -> str:
    """create HTML to display an icon next to a name. Can also be a link."""
    name_html = link_html(url, name, new_window=False) if url else name
    if text:
        name_html = format_html("{}&nbsp;{}", name_html, text)

    return format_html(
        "{}&nbsp;&nbsp;&nbsp;{}",
        create_img_html(
            icon_url, classes=["ra-avatar", "img-circle"] if avatar else [], size=size
        ),
        name_html,
    )


@login_required
@permission_required("standingssync.add_syncedcharacter")
def index(request):
    """main page"""
    if not request.user.profile.main_character:
        sync_manager = None
    else:
        try:
            alliance = EveAllianceInfo.objects.get(
                alliance_id=request.user.profile.main_character.alliance_id
            )
            sync_manager = SyncManager.objects.get(alliance=alliance)
        except EveAllianceInfo.DoesNotExist:
            sync_manager = None
        except SyncManager.DoesNotExist:
            sync_manager = None

    # get list of synced synced_characters for this user
    characters_query = SyncedCharacter.objects.select_related(
        "character_ownership__character"
    ).filter(character_ownership__user=request.user)
    synced_characters = list()
    for synced_character in characters_query:
        character = synced_character.character_ownership.character
        name_html = create_icon_plus_name_html(
            character.portrait_url(), character.character_name, avatar=True
        )
        organization = character.corporation_name
        if character.alliance_ticker:
            organization += f" [{character.alliance_ticker}]"

        status_message_raw = synced_character.get_status_message()
        if status_message_raw.lower() == "ok":
            if (
                STANDINGSSYNC_ADD_WAR_TARGETS
                and synced_character.has_war_targets_label is False
            ):
                status_message = format_html(
                    '<span class="text-warning">'
                    '<i class="fas fa-info-circle"></i> '
                    "Please create a contact label with the name: {}</span>",
                    STANDINGSSYNC_WAR_TARGETS_LABEL_NAME,
                )
            else:
                status_message = format_html(
                    '<i class="fas fa-check text-success"></i>'
                )
        else:
            status_message = status_message_raw
            status_message = format_html(
                '<span class="text-danger">'
                '<i class="fas fa-exclamation-triangle"></i> {}</span>',
                status_message_raw,
            )

        synced_characters.append(
            {
                "name": character.character_name,
                "name_html": name_html,
                "organization": organization,
                "status_message": status_message,
                "has_error": synced_character.last_error != SyncedCharacter.Error.NONE,
                "pk": synced_character.pk,
            }
        )

    context = {
        "app_title": __title__,
        "synced_characters": synced_characters,
        "has_synced_chars": len(synced_characters) > 0,
    }
    if sync_manager:
        context["alliance"] = sync_manager.alliance
        if STANDINGSSYNC_REPLACE_CONTACTS:
            alliance_contacts_count = sync_manager.contacts.filter(
                is_war_target=False
            ).count()
        else:
            alliance_contacts_count = None
        if STANDINGSSYNC_ADD_WAR_TARGETS:
            alliance_war_targets_count = sync_manager.contacts.filter(
                is_war_target=True
            ).count()
        else:
            alliance_war_targets_count = None
    else:
        context["alliance"] = None
        alliance_contacts_count = None
        alliance_war_targets_count = None

    context["alliance_contacts_count"] = alliance_contacts_count
    context["alliance_war_targets_count"] = alliance_war_targets_count
    context["war_targets_label_name"] = STANDINGSSYNC_WAR_TARGETS_LABEL_NAME

    return render(request, "standingssync/index.html", context)


@login_required
@permission_required("standingssync.add_syncmanager")
@token_required(SyncManager.get_esi_scopes())
def add_alliance_manager(request, token):
    """add or update sync manager for an alliance"""
    success = True
    token_char = EveCharacter.objects.get(character_id=token.character_id)
    character_ownership = None
    alliance = None

    if not token_char.alliance_id:
        messages_plus.warning(
            request,
            f"Can not add {token_char}, because it is not a member of any alliance.",
        )
        success = False

    if success:
        try:
            character_ownership = CharacterOwnership.objects.get(
                user=request.user, character=token_char
            )
        except CharacterOwnership.DoesNotExist:
            messages_plus.warning(
                request, f"Could not find character {token_char.character_name}"
            )
            success = False

    if success:
        try:
            alliance = EveAllianceInfo.objects.get(alliance_id=token_char.alliance_id)
        except EveAllianceInfo.DoesNotExist:
            alliance = EveAllianceInfo.objects.create_alliance(token_char.alliance_id)
            alliance.save()

    if success:
        sync_manager, _ = SyncManager.objects.update_or_create(
            alliance=alliance, defaults={"character_ownership": character_ownership}
        )
        tasks.run_manager_sync.delay(sync_manager.pk)
        messages_plus.success(
            request,
            "{} set as alliance character for {}. "
            "Started syncing of alliance contacts. ".format(
                sync_manager.character_ownership.character.character_name,
                alliance.alliance_name,
            ),
        )
    return redirect("standingssync:index")


@login_required
@permission_required("standingssync.add_syncedcharacter")
@token_required(scopes=SyncedCharacter.get_esi_scopes())
def add_character(request, token):
    """add character to receive alliance contacts"""
    try:
        alliance = EveAllianceInfo.objects.get(
            alliance_id=request.user.profile.main_character.alliance_id
        )
    except EveAllianceInfo.DoesNotExist:
        raise RuntimeError("Can not find alliance")

    try:
        sync_manager = SyncManager.objects.get(alliance=alliance)
    except SyncManager.DoesNotExist:
        raise RuntimeError("can not find sync manager for alliance")

    token_char = EveCharacter.objects.get(character_id=token.character_id)
    if token_char.alliance_id == sync_manager.character_ownership.character.alliance_id:
        messages_plus.warning(
            request,
            "Adding alliance members does not make much sense, "
            "since they already have access to alliance contacts.",
        )

    else:
        try:
            character_ownership = CharacterOwnership.objects.get(
                user=request.user, character=token_char
            )
        except CharacterOwnership.DoesNotExist:
            messages_plus.warning(
                request, "Could not find character {}".format(token_char.character_name)
            )
        else:
            eff_standing = sync_manager.get_effective_standing(
                character_ownership.character
            )
            if eff_standing < STANDINGSSYNC_CHAR_MIN_STANDING:
                messages_plus.warning(
                    request,
                    "Can not activate sync for your "
                    f"character {token_char.character_name}, "
                    "because it does not have blue standing "
                    "with the alliance. "
                    f"The standing value is: {eff_standing:.1f}. "
                    "Please first obtain blue "
                    "standing for your character and then try again.",
                )
            else:
                sync_character, _ = SyncedCharacter.objects.update_or_create(
                    character_ownership=character_ownership,
                    defaults={"manager": sync_manager},
                )
                tasks.run_character_sync.delay(sync_character.pk)
                messages_plus.success(
                    request, "Sync activated for {}!".format(token_char.character_name)
                )
    return redirect("standingssync:index")


@login_required
@permission_required("standingssync.add_syncedcharacter")
def remove_character(request, alt_pk):
    """remove character from receiving alliance contacts"""
    alt = SyncedCharacter.objects.get(pk=alt_pk)
    alt_name = alt.character_ownership.character.character_name
    alt.delete()
    messages_plus.success(request, "Sync deactivated for {}".format(alt_name))
    return redirect("standingssync:index")



import datetime as dt
from unittest.mock import Mock, patch

from django.test import TestCase
from django.utils.timezone import now
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.esi_testing import BravadoOperationStub
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from ..models import EveContact, EveWar, SyncedCharacter, SyncManager
from .factories import (
    EveContactFactory,
    EveContactWarTargetFactory,
    EveEntityAllianceFactory,
    EveWarFactory,
    SyncedCharacterFactory,
    SyncManagerFactory,
    UserMainSyncerFactory,
)
from .utils import (
    ALLIANCE_CONTACTS,
    EsiCharacterContactsStub,
    EsiContact,
    EsiContactLabel,
    LoadTestDataMixin,
)

MODELS_PATH = "standingssync.models"


class TestGetEffectiveStanding(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        user, _ = create_user_from_evecharacter(
            cls.character_1.character_id, permissions=["standingssync.add_syncmanager"]
        )
        cls.sync_manager = SyncManagerFactory(user=user)
        contacts = [
            {"contact_id": 1001, "contact_type": "character", "standing": -10},
            {"contact_id": 2001, "contact_type": "corporation", "standing": 10},
            {"contact_id": 3001, "contact_type": "alliance", "standing": 5},
        ]
        for contact in contacts:
            EveContactFactory(
                manager=cls.sync_manager,
                eve_entity=EveEntity.objects.get(id=contact["contact_id"]),
                standing=contact["standing"],
            )

    def test_char_with_character_standing(self):
        c1 = EveCharacter(
            character_id=1001,
            character_name="Char 1",
            corporation_id=201,
            corporation_name="Corporation 1",
            corporation_ticker="C1",
        )
        self.assertEqual(self.sync_manager.get_effective_standing(c1), -10)

    def test_char_with_corporation_standing(self):
        c2 = EveCharacter(
            character_id=1002,
            character_name="Char 2",
            corporation_id=2001,
            corporation_name="Corporation 1",
            corporation_ticker="C1",
        )
        self.assertEqual(self.sync_manager.get_effective_standing(c2), 10)

    def test_char_with_alliance_standing(self):
        c3 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
            alliance_id=3001,
            alliance_name="Alliance 1",
            alliance_ticker="A1",
        )
        self.assertEqual(self.sync_manager.get_effective_standing(c3), 5)

    def test_char_without_standing_and_has_alliance(self):
        c4 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
            alliance_id=3002,
            alliance_name="Alliance 2",
            alliance_ticker="A2",
        )
        self.assertEqual(self.sync_manager.get_effective_standing(c4), 0.0)

    def test_char_without_standing_and_without_alliance_1(self):
        c4 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
            alliance_id=None,
            alliance_name=None,
            alliance_ticker=None,
        )
        self.assertEqual(self.sync_manager.get_effective_standing(c4), 0.0)

    def test_char_without_standing_and_without_alliance_2(self):
        c4 = EveCharacter(
            character_id=1003,
            character_name="Char 3",
            corporation_id=2003,
            corporation_name="Corporation 3",
            corporation_ticker="C2",
        )
        self.assertEqual(self.sync_manager.get_effective_standing(c4), 0.0)


class TestSyncManager(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # 1 user has permission for manager sync
        cls.user_1, _ = create_user_from_evecharacter(
            cls.character_1.character_id, permissions=["standingssync.add_syncmanager"]
        )

        # user 1 has no permission for manager sync and has 1 alt
        cls.user_2, _ = create_user_from_evecharacter(cls.character_2.character_id)
        cls.alt_ownership = CharacterOwnership.objects.create(
            character=cls.character_4, owner_hash="x4", user=cls.user_2
        )

    def test_should_report_no_sync_error(self):
        # given
        sync_manager = SyncManagerFactory(user=self.user_1)
        sync_manager.set_sync_status(SyncManager.Error.NONE)
        # when/then
        self.assertTrue(sync_manager.is_sync_ok)

    def test_should_report_sync_error(self):
        # given
        sync_manager = SyncManagerFactory(user=self.user_1)
        for status in [
            SyncManager.Error.TOKEN_INVALID,
            SyncManager.Error.TOKEN_EXPIRED,
            SyncManager.Error.INSUFFICIENT_PERMISSIONS,
            SyncManager.Error.NO_CHARACTER,
            SyncManager.Error.ESI_UNAVAILABLE,
            SyncManager.Error.UNKNOWN,
        ]:
            with self.subTest(status=status):
                sync_manager.set_sync_status(status)
                # when/then
                self.assertFalse(sync_manager.is_sync_ok)

    def test_set_sync_status(self):
        # given
        sync_manager = SyncManagerFactory(user=self.user_1)
        sync_manager.last_error = SyncManager.Error.NONE
        sync_manager.last_sync = None
        # when
        sync_manager.set_sync_status(SyncManager.Error.TOKEN_INVALID)
        # then
        sync_manager.refresh_from_db()
        self.assertEqual(sync_manager.last_error, SyncManager.Error.TOKEN_INVALID)
        self.assertIsNotNone(sync_manager.last_sync)

    def test_should_abort_when_no_char(self):
        # given
        sync_manager = SyncManagerFactory(
            alliance=self.alliance_1, character_ownership=None
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertFalse(result)
        sync_manager.refresh_from_db()
        self.assertEqual(sync_manager.last_error, SyncManager.Error.NO_CHARACTER)

    def test_should_abort_when_insufficient_permission(self):
        # given
        sync_manager = SyncManagerFactory(user=self.user_2)
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertFalse(result)
        sync_manager.refresh_from_db()
        self.assertEqual(
            sync_manager.last_error, SyncManager.Error.INSUFFICIENT_PERMISSIONS
        )

    @patch(MODELS_PATH + ".Token")
    def test_should_report_error_when_character_has_no_token(self, mock_Token):
        # given
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = (
            None
        )
        sync_manager = SyncManagerFactory(user=self.user_1)
        # when
        result = sync_manager.update_from_esi()
        # then
        sync_manager.refresh_from_db()
        self.assertFalse(result)
        self.assertEqual(sync_manager.last_error, SyncManager.Error.TOKEN_INVALID)

    @patch(MODELS_PATH + ".Token")
    def test_should_report_error_when_token_is_expired(self, mock_Token):
        # given
        mock_Token.objects.filter.side_effect = TokenExpiredError()
        sync_manager = SyncManagerFactory(user=self.user_1)
        SyncedCharacterFactory(
            character_ownership=self.alt_ownership, manager=sync_manager
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        sync_manager.refresh_from_db()
        self.assertFalse(result)
        self.assertEqual(sync_manager.last_error, SyncManager.Error.TOKEN_EXPIRED)

    @patch(MODELS_PATH + ".Token")
    def test_should_report_error_when_token_is_invalid(self, mock_Token):
        # given
        mock_Token.objects.filter.side_effect = TokenInvalidError()
        sync_manager = SyncManagerFactory(user=self.user_1)
        SyncedCharacterFactory(
            character_ownership=self.alt_ownership, manager=sync_manager
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        sync_manager.refresh_from_db()
        self.assertFalse(result)
        self.assertEqual(sync_manager.last_error, SyncManager.Error.TOKEN_INVALID)

    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_sync_contacts(self, mock_esi, mock_Token):
        # given
        sync_manager = SyncManagerFactory(user=self.user_1)
        SyncedCharacterFactory(
            character_ownership=self.alt_ownership, manager=sync_manager
        )
        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False):
            # when
            self._run_sync(sync_manager, mock_esi, mock_Token)
        # then (continued)
        contact = sync_manager.contacts.get(eve_entity_id=3015)
        self.assertEqual(contact.standing, 10.0)
        self.assertFalse(contact.is_war_target)

    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_sync_contacts_and_war_targets(self, mock_esi, mock_Token):
        # given
        sync_manager = SyncManagerFactory(user=self.user_1)
        SyncedCharacterFactory(
            character_ownership=self.alt_ownership, manager=sync_manager
        )
        EveWar.objects.create(
            id=8,
            aggressor=EveEntity.objects.get(id=3015),
            defender=EveEntity.objects.get(id=3001),
            declared=now() - dt.timedelta(days=3),
            started=now() - dt.timedelta(days=2),
            is_mutual=False,
            is_open_for_allies=False,
        )

        with patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True):
            # when
            self._run_sync(sync_manager, mock_esi, mock_Token)
        # then (continued)
        contact = sync_manager.contacts.get(eve_entity_id=3015)
        self.assertEqual(contact.standing, -10.0)
        self.assertTrue(contact.is_war_target)

    def _run_sync(self, sync_manager, mock_esi, mock_Token):
        def esi_get_alliances_alliance_id_contacts(*args, **kwargs):
            return BravadoOperationStub(ALLIANCE_CONTACTS)

        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.side_effect = (
            esi_get_alliances_alliance_id_contacts
        )
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = Mock(
            spec=Token
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertEqual(sync_manager.last_error, SyncManager.Error.NONE)
        expected_contact_ids = {x["contact_id"] for x in ALLIANCE_CONTACTS}
        expected_contact_ids.add(self.character_1.alliance_id)
        result_contact_ids = set(
            sync_manager.contacts.values_list("eve_entity_id", flat=True)
        )
        self.assertSetEqual(expected_contact_ids, result_contact_ids)
        return sync_manager


def fetch_war_targets():
    return set(
        EveContact.objects.filter(is_war_target=True).values_list(
            "eve_entity_id", flat=True
        )
    )


@patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
@patch(MODELS_PATH + ".esi")
class TestSyncManager2(NoSocketsTestCase):
    def test_should_add_war_target_contact_as_aggressor_1(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            aggressor=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id)
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertSetEqual(fetch_war_targets(), {war.defender.id})

    def test_should_add_war_target_contact_as_aggressor_2(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        sync_manager = SyncManagerFactory()
        ally = EveEntityAllianceFactory()
        war = EveWarFactory(
            aggressor=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            allies=[ally],
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertSetEqual(fetch_war_targets(), {war.defender.id, ally.id})

    def test_should_add_war_target_contact_as_defender(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id)
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertSetEqual(fetch_war_targets(), {war.aggressor.id})

    def test_should_add_war_target_contact_as_ally(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            allies=[EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id)]
        )
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertSetEqual(fetch_war_targets(), {war.aggressor.id})

    def test_should_not_add_war_target_contact_from_unrelated_war(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        sync_manager = SyncManagerFactory()
        EveWarFactory()
        EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id)
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertSetEqual(fetch_war_targets(), set())

    def test_remove_outdated_war_target_contacts(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        sync_manager = SyncManagerFactory()
        war = EveWarFactory(
            defender=EveEntityAllianceFactory(id=sync_manager.alliance.alliance_id),
            finished=now(),
        )
        EveContactWarTargetFactory(manager=sync_manager, eve_entity=war.aggressor)
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)
        sync_manager.refresh_from_db()
        self.assertSetEqual(fetch_war_targets(), set())

    def test_do_nothing_when_contacts_are_unchanged(self, mock_esi):
        # given
        mock_esi.client.Contacts.get_alliances_alliance_id_contacts.return_value = (
            BravadoOperationStub([])
        )
        my_version_hash = SyncManager._calculate_version_hash({})
        sync_manager = SyncManagerFactory(version_hash=my_version_hash)
        # when
        result = sync_manager.update_from_esi()
        # then
        self.assertTrue(result)


@patch(MODELS_PATH + ".STANDINGSSYNC_WAR_TARGETS_LABEL_NAME", "WAR TARGETS")
class TestSyncCharacter(LoadTestDataMixin, TestCase):
    CHARACTER_CONTACTS = [
        EsiContact(1014, EsiContact.ContactType.CHARACTER, standing=10.0),
        EsiContact(2011, EsiContact.ContactType.CORPORATION, standing=5.0),
        EsiContact(3011, EsiContact.ContactType.ALLIANCE, standing=-10.0),
    ]

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # 1 user with 1 alt character
        cls.user_1, _ = create_user_from_evecharacter(cls.character_1.character_id)
        cls.alt_ownership_2 = CharacterOwnership.objects.create(
            character=cls.character_2, owner_hash="x2", user=cls.user_1
        )
        cls.alt_ownership_3 = CharacterOwnership.objects.create(
            character=cls.character_3, owner_hash="x3", user=cls.user_1
        )
        cls.sync_manager = SyncManagerFactory(user=cls.user_1, version_hash="new")
        # sync manager with contacts
        for contact in ALLIANCE_CONTACTS:
            EveContactFactory(
                manager=cls.sync_manager,
                eve_entity=EveEntity.objects.get(id=contact["contact_id"]),
                standing=contact["standing"],
            )
        # set to contacts as war targets
        cls.sync_manager.contacts.filter(eve_entity_id__in=[1014, 3013]).update(
            is_war_target=True, standing=-10.0
        )
        cls.alliance_contacts = [
            cls.eve_contact_2_esi_contact(obj)
            for obj in cls.sync_manager.contacts.all()
        ]

    @staticmethod
    def eve_contact_2_esi_contact(eve_contact):
        map_category_2_type = {
            EveEntity.CATEGORY_CHARACTER: EsiContact.ContactType.CHARACTER,
            EveEntity.CATEGORY_CORPORATION: EsiContact.ContactType.CORPORATION,
            EveEntity.CATEGORY_ALLIANCE: EsiContact.ContactType.ALLIANCE,
        }
        return EsiContact(
            contact_id=eve_contact.eve_entity_id,
            contact_type=map_category_2_type[eve_contact.eve_entity.category],
            standing=eve_contact.standing,
        )

    def setUp(self) -> None:
        self.maxDiff = None
        self.synced_character_2 = SyncedCharacterFactory(
            character_ownership=self.alt_ownership_2, manager=self.sync_manager
        )
        self.synced_character_3 = SyncedCharacterFactory(
            character_ownership=self.alt_ownership_3, manager=self.sync_manager
        )

    def test_should_report_no_sync_error(self):
        # given
        self.synced_character_2.set_sync_status(SyncManager.Error.NONE)
        # when/then
        self.assertTrue(self.synced_character_2.is_sync_ok)

    def test_should_report_sync_error(self):
        # given
        for status in [
            SyncedCharacter.Error.TOKEN_INVALID,
            SyncedCharacter.Error.TOKEN_EXPIRED,
            SyncedCharacter.Error.INSUFFICIENT_PERMISSIONS,
            SyncedCharacter.Error.ESI_UNAVAILABLE,
            SyncedCharacter.Error.UNKNOWN,
        ]:
            self.synced_character_2.set_sync_status(status)
            # when/then
            self.assertFalse(self.synced_character_2.is_sync_ok)

    def test_get_last_error_message_after_sync(self):
        self.synced_character_2.last_sync = now()
        self.synced_character_2.last_error = SyncedCharacter.Error.NONE
        expected = "OK"
        self.assertEqual(self.synced_character_2.get_status_message(), expected)

        self.synced_character_2.last_error = SyncedCharacter.Error.TOKEN_EXPIRED
        expected = "Expired token"
        self.assertEqual(self.synced_character_2.get_status_message(), expected)

    def test_get_last_error_message_no_sync(self):
        self.synced_character_2.last_sync = None
        self.synced_character_2.last_error = SyncedCharacter.Error.NONE
        expected = "Not synced yet"
        self.assertEqual(self.synced_character_2.get_status_message(), expected)

        self.synced_character_2.last_error = SyncedCharacter.Error.TOKEN_EXPIRED
        expected = "Expired token"
        self.assertEqual(self.synced_character_2.get_status_message(), expected)

    def test_set_sync_status(self):
        self.synced_character_2.last_error = SyncManager.Error.NONE
        self.synced_character_2.last_sync = None

        self.synced_character_2.set_sync_status(SyncManager.Error.TOKEN_INVALID)
        self.synced_character_2.refresh_from_db()

        self.assertEqual(
            self.synced_character_2.last_error, SyncManager.Error.TOKEN_INVALID
        )
        self.assertIsNotNone(self.synced_character_2.last_sync)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_do_nothing_if_no_update_needed(self, mock_esi, mock_Token):
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        self.synced_character_2.version_hash = self.sync_manager.version_hash
        self.synced_character_2.save()
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertIsNone(self.synced_character_2.last_sync)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_replace_all_contacts_1(self, mock_esi, mock_Token):
        """run normal sync for a character which has blue standing"""
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertEqual(self.synced_character_2.last_error, SyncedCharacter.Error.NONE)
        alliance_contacts_2 = {
            obj for obj in self.alliance_contacts if obj.contact_id != character_id
        }
        self.assertSetEqual(
            set(esi_character_contacts.contacts(character_id)), alliance_contacts_2
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.0)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_replace_all_contacts_2(self, mock_esi, mock_Token):
        """run normal sync for a character which has no standing and allow neutrals"""
        # given
        character_id = (
            self.synced_character_3.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_3, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertEqual(self.synced_character_3.last_error, SyncedCharacter.Error.NONE)
        self.assertSetEqual(
            set(esi_character_contacts.contacts(character_id)),
            set(self.alliance_contacts),
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_replace_all_contacts_and_add_war_targets(
        self, mock_esi, mock_Token
    ):
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_labels(
            character_id, [EsiContactLabel(1, "war targets")]
        )
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertEqual(self.synced_character_2.last_error, SyncedCharacter.Error.NONE)
        result = set(esi_character_contacts.contacts(character_id))
        expected = {
            EsiContact(
                1014,
                EsiContact.ContactType.CHARACTER,
                standing=-10.0,
                label_ids=[1],
            ),
            EsiContact(3011, EsiContact.ContactType.ALLIANCE, standing=-10.0),
            EsiContact(
                3013, EsiContact.ContactType.ALLIANCE, standing=-10.0, label_ids=[1]
            ),
            EsiContact(1016, EsiContact.ContactType.CHARACTER, standing=10.0),
            EsiContact(2013, EsiContact.ContactType.CORPORATION, standing=5.0),
            EsiContact(2012, EsiContact.ContactType.CORPORATION, standing=-5.0),
            EsiContact(1005, EsiContact.ContactType.CHARACTER, standing=-10.0),
            EsiContact(1013, EsiContact.ContactType.CHARACTER, standing=-5.0),
            EsiContact(3014, EsiContact.ContactType.ALLIANCE, standing=5.0),
            EsiContact(2015, EsiContact.ContactType.CORPORATION, standing=10.0),
            EsiContact(2011, EsiContact.ContactType.CORPORATION, standing=-10.0),
            EsiContact(2014, EsiContact.ContactType.CORPORATION, standing=0.0),
            EsiContact(3015, EsiContact.ContactType.ALLIANCE, standing=10.0),
            EsiContact(1012, EsiContact.ContactType.CHARACTER, standing=-10.0),
            EsiContact(1015, EsiContact.ContactType.CHARACTER, standing=5.0),
            EsiContact(1004, EsiContact.ContactType.CHARACTER, standing=10.0),
            EsiContact(3012, EsiContact.ContactType.ALLIANCE, standing=-5.0),
        }
        self.assertSetEqual(result, expected)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_update_war_targets_only_1(self, mock_esi, mock_Token):
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertEqual(self.synced_character_2.last_error, SyncedCharacter.Error.NONE)
        expected = {
            EsiContact(1014, EsiContact.ContactType.CHARACTER, standing=-10.0),
            EsiContact(2011, EsiContact.ContactType.CORPORATION, standing=5.0),
            EsiContact(3011, EsiContact.ContactType.ALLIANCE, standing=-10.0),
            EsiContact(3013, EsiContact.ContactType.ALLIANCE, standing=-10.0),
        }
        self.assertSetEqual(
            set(esi_character_contacts.contacts(character_id)),
            expected,
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_update_war_targets_only_2(self, mock_esi, mock_Token):
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        war_target_label = EsiContactLabel(99, "war targets")
        esi_character_contacts.setup_labels(
            character_id, [war_target_label, EsiContactLabel(1, "other")]
        )
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertEqual(self.synced_character_2.last_error, SyncedCharacter.Error.NONE)
        expected = {
            EsiContact(
                1014,
                EsiContact.ContactType.CHARACTER,
                standing=-10.0,
                label_ids=[war_target_label.id],
            ),
            EsiContact(2011, EsiContact.ContactType.CORPORATION, standing=5.0),
            EsiContact(3011, EsiContact.ContactType.ALLIANCE, standing=-10.0),
            EsiContact(
                3013,
                EsiContact.ContactType.ALLIANCE,
                standing=-10.0,
                label_ids=[war_target_label.id],
            ),
        }
        self.assertSetEqual(
            set(esi_character_contacts.contacts(character_id)),
            expected,
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_remove_outdated_war_targets_with_label(self, mock_esi, mock_Token):
        # given
        character_id = self.synced_character_2.character.character_id
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_labels(
            character_id, [EsiContactLabel(1, "war targets")]
        )
        esi_character_contacts.setup_contacts(
            character_id,
            [
                EsiContact(
                    1011,
                    EsiContact.ContactType.CHARACTER,
                    standing=-10.0,
                    label_ids=[1],
                ),
                EsiContact(1014, EsiContact.ContactType.CHARACTER, standing=10.0),
                EsiContact(2011, EsiContact.ContactType.CORPORATION, standing=5.0),
                EsiContact(3011, EsiContact.ContactType.ALLIANCE, standing=-10.0),
            ],
        )
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertEqual(self.synced_character_2.last_error, SyncedCharacter.Error.NONE)
        expected = {
            EsiContact(
                1014,
                EsiContact.ContactType.CHARACTER,
                standing=-10.0,
                label_ids=[1],
            ),
            EsiContact(2011, EsiContact.ContactType.CORPORATION, standing=5.0),
            EsiContact(3011, EsiContact.ContactType.ALLIANCE, standing=-10.0),
            EsiContact(
                3013, EsiContact.ContactType.ALLIANCE, standing=-10.0, label_ids=[1]
            ),
        }
        self.assertSetEqual(
            set(esi_character_contacts.contacts(character_id)),
            expected,
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_record_if_character_has_wt_label(self, mock_esi, mock_Token):
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_labels(
            character_id, [EsiContactLabel(1, "war targets")]
        )
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertTrue(self.synced_character_2.has_war_targets_label)

    @patch(MODELS_PATH + ".STANDINGSSYNC_ADD_WAR_TARGETS", True)
    @patch(MODELS_PATH + ".STANDINGSSYNC_REPLACE_CONTACTS", False)
    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.01)
    @patch(MODELS_PATH + ".Token")
    @patch(MODELS_PATH + ".esi")
    def test_should_record_if_character_does_not_have_wt_label(
        self, mock_esi, mock_Token
    ):
        # given
        character_id = (
            self.synced_character_2.character_ownership.character.character_id
        )
        esi_character_contacts = EsiCharacterContactsStub()
        esi_character_contacts.setup_contacts(character_id, self.CHARACTER_CONTACTS)
        # when
        result = self._run_sync(
            mock_esi, mock_Token, self.synced_character_2, esi_character_contacts
        )
        # then
        self.assertTrue(result)
        self.assertFalse(self.synced_character_2.has_war_targets_label)

    @staticmethod
    def _run_sync(mock_esi, mock_Token, synced_character, esi_character_contacts):
        # given
        esi_character_contacts.setup_esi_mock(mock_esi)
        mock_Token.objects.filter = Mock()
        synced_character.character_ownership.user = (
            AuthUtils.add_permission_to_user_by_name(
                "standingssync.add_syncedcharacter",
                synced_character.character_ownership.user,
            )
        )
        # when
        result = synced_character.update()
        if result:
            synced_character.refresh_from_db()
        return result

    def test_should_deactivate_when_insufficient_permission(self):
        # when
        result = self.synced_character_2.update()
        # then
        self.assertFalse(result)
        self.assertFalse(
            SyncedCharacter.objects.filter(pk=self.synced_character_2.pk).exists()
        )

    @patch(MODELS_PATH + ".Token")
    def test_should_deactivate_when_no_token_found(self, mock_Token):
        # given
        mock_Token.objects.filter.return_value = Token.objects.none()
        self.synced_character_2.character_ownership.user = (
            AuthUtils.add_permission_to_user_by_name(
                "standingssync.add_syncedcharacter",
                self.synced_character_2.character_ownership.user,
            )
        )
        # when
        result = self.synced_character_2.update()
        # then
        self.assertFalse(result)
        self.assertFalse(
            SyncedCharacter.objects.filter(pk=self.synced_character_2.pk).exists()
        )

    @patch(MODELS_PATH + ".Token")
    def test_should_deactivate_when_token_is_invalid(self, mock_Token):
        # given
        mock_Token.objects.filter.side_effect = TokenInvalidError()
        self.synced_character_2.character_ownership.user = (
            AuthUtils.add_permission_to_user_by_name(
                "standingssync.add_syncedcharacter",
                self.synced_character_2.character_ownership.user,
            )
        )
        # when
        result = self.synced_character_2.update()
        # then
        self.assertFalse(result)
        self.assertFalse(
            SyncedCharacter.objects.filter(pk=self.synced_character_2.pk).exists()
        )

    @patch(MODELS_PATH + ".Token")
    def test_should_deactivate_when_token_is_expired(self, mock_Token):
        # given
        mock_Token.objects.filter.side_effect = TokenExpiredError()
        self.synced_character_2.character_ownership.user = (
            AuthUtils.add_permission_to_user_by_name(
                "standingssync.add_syncedcharacter",
                self.synced_character_2.character_ownership.user,
            )
        )
        # when
        result = self.synced_character_2.update()
        # then
        self.assertFalse(result)
        self.assertFalse(
            SyncedCharacter.objects.filter(pk=self.synced_character_2.pk).exists()
        )

    @patch(MODELS_PATH + ".STANDINGSSYNC_CHAR_MIN_STANDING", 0.1)
    @patch(MODELS_PATH + ".Token")
    def test_should_deactivate_when_character_has_no_standing(self, mock_Token):
        # given
        mock_Token.objects.filter.return_value = Mock()
        self.synced_character_2.character_ownership.user = (
            AuthUtils.add_permission_to_user_by_name(
                "standingssync.add_syncedcharacter",
                self.synced_character_2.character_ownership.user,
            )
        )
        contact = self.sync_manager.contacts.get(
            eve_entity_id=self.character_2.character_id
        )
        contact.standing = -10
        contact.save()
        # when
        result = self.synced_character_2.update()
        # then
        self.assertFalse(result)
        self.assertFalse(
            SyncedCharacter.objects.filter(pk=self.synced_character_2.pk).exists()
        )


class TestSyncCharacter2(NoSocketsTestCase):
    def test_should_not_sync_when_no_contacts(self):
        # given
        manager = SyncManagerFactory(version_hash="abc")
        character = SyncedCharacterFactory(manager=manager)
        # when
        result = character.update()
        # then
        self.assertTrue(result)

    def test_should_abort_sync_when_insufficient_permissions(self):
        # given
        manager = SyncManagerFactory(version_hash="abc")
        user = UserMainSyncerFactory(permissions__=[])
        character = SyncedCharacterFactory(manager=manager, user=user)
        # when
        result = character.update()
        # then
        self.assertFalse(result)


class TestEveWar(NoSocketsTestCase):
    def test_str(self):
        # given
        aggressor = EveEntityAllianceFactory(name="Alpha")
        defender = EveEntityAllianceFactory(name="Bravo")
        war = EveWarFactory(aggressor=aggressor, defender=defender)
        # when/then
        self.assertEqual(str(war), "Alpha vs. Bravo")



# Generated by Django 4.0.6 on 2022-07-29 17:14

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0007_evetype_description"),
        ("standingssync", "0004_remove_old_eve_entity"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveWar",
            fields=[
                ("id", models.PositiveIntegerField(primary_key=True, serialize=False)),
                ("declared", models.DateTimeField()),
                (
                    "finished",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                ("is_mutual", models.BooleanField()),
                ("is_open_for_allies", models.BooleanField()),
                ("retracted", models.DateTimeField(default=None, null=True)),
                (
                    "started",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                (
                    "aggressor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "allies",
                    models.ManyToManyField(
                        related_name="+", to="eveuniverse.eveentity"
                    ),
                ),
                (
                    "defender",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveContact",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("standing", models.FloatField()),
                ("is_war_target", models.BooleanField()),
                (
                    "eve_entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "manager",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="contacts",
                        to="standingssync.syncmanager",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="evecontact",
            constraint=models.UniqueConstraint(
                fields=("manager", "eve_entity"), name="fk_eve_contact"
            ),
        ),
    ]



# Generated by Django 3.1.4 on 2021-01-12 00:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveonline", "0012_index_additions"),
        ("standingssync", "0001_initial-new"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveContact",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("standing", models.FloatField()),
                ("is_war_target", models.BooleanField()),
            ],
        ),
        migrations.CreateModel(
            name="EveEntity",
            fields=[
                ("id", models.PositiveIntegerField(primary_key=True, serialize=False)),
                (
                    "category",
                    models.CharField(
                        choices=[
                            ("AL", "alliance"),
                            ("CO", "corporation"),
                            ("CH", "character"),
                        ],
                        db_index=True,
                        max_length=2,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveWar",
            fields=[
                ("id", models.PositiveIntegerField(primary_key=True, serialize=False)),
                ("declared", models.DateTimeField()),
                (
                    "finished",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                ("is_mutual", models.BooleanField()),
                ("is_open_for_allies", models.BooleanField()),
                ("retracted", models.DateTimeField(default=None, null=True)),
                (
                    "started",
                    models.DateTimeField(db_index=True, default=None, null=True),
                ),
                (
                    "aggressor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="aggressor_war",
                        to="standingssync.eveentity",
                    ),
                ),
                (
                    "allies",
                    models.ManyToManyField(
                        related_name="ally", to="standingssync.EveEntity"
                    ),
                ),
                (
                    "defender",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="defender_war",
                        to="standingssync.eveentity",
                    ),
                ),
            ],
        ),
        migrations.RenameField(
            model_name="syncedcharacter",
            old_name="character",
            new_name="character_ownership",
        ),
        migrations.RenameField(
            model_name="syncmanager",
            old_name="character",
            new_name="character_ownership",
        ),
        migrations.AlterField(
            model_name="syncedcharacter",
            name="manager",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="synced_characters",
                to="standingssync.syncmanager",
            ),
        ),
        migrations.AlterField(
            model_name="syncedcharacter",
            name="version_hash",
            field=models.CharField(default="", max_length=32),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="alliance",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                related_name="+",
                serialize=False,
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="version_hash",
            field=models.CharField(default="", max_length=32),
        ),
        migrations.DeleteModel(
            name="AllianceContact",
        ),
        migrations.AddField(
            model_name="evecontact",
            name="eve_entity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contacts",
                to="standingssync.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="evecontact",
            name="manager",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contacts",
                to="standingssync.syncmanager",
            ),
        ),
        migrations.AddConstraint(
            model_name="evecontact",
            constraint=models.UniqueConstraint(
                fields=("manager", "eve_entity"), name="fk_eve_contact"
            ),
        ),
    ]



"""Utility functions and classes for tests"""
import copy
from dataclasses import dataclass, field
from enum import Enum
from typing import FrozenSet, List

from eveuniverse.models import EveEntity

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from app_utils.esi_testing import BravadoOperationStub


def create_esi_contact(eve_entity: EveEntity, standing: int = 5.0) -> dict:
    if standing < -10 or standing > 10:
        raise ValueError(f"Invalid standing: {standing}")
    params = {
        "contact_id": int(eve_entity.id),
        "contact_type": eve_entity.category,
        "standing": float(standing),
    }
    return params


ALLIANCE_CONTACTS = [
    {"contact_id": 1002, "contact_type": "character", "standing": 10.0},
    {"contact_id": 1004, "contact_type": "character", "standing": 10.0},
    {"contact_id": 1005, "contact_type": "character", "standing": -10.0},
    {"contact_id": 1012, "contact_type": "character", "standing": -10.0},
    {"contact_id": 1013, "contact_type": "character", "standing": -5.0},
    {"contact_id": 1014, "contact_type": "character", "standing": 0.0},
    {"contact_id": 1015, "contact_type": "character", "standing": 5.0},
    {"contact_id": 1016, "contact_type": "character", "standing": 10.0},
    {"contact_id": 3011, "contact_type": "alliance", "standing": -10.0},
    {"contact_id": 3012, "contact_type": "alliance", "standing": -5.0},
    {"contact_id": 3013, "contact_type": "alliance", "standing": 0.0},
    {"contact_id": 3014, "contact_type": "alliance", "standing": 5.0},
    {"contact_id": 3015, "contact_type": "alliance", "standing": 10.0},
    {"contact_id": 2011, "contact_type": "corporation", "standing": -10.0},
    {"contact_id": 2012, "contact_type": "corporation", "standing": -5.0},
    {"contact_id": 2014, "contact_type": "corporation", "standing": 0.0},
    {"contact_id": 2013, "contact_type": "corporation", "standing": 5.0},
    {"contact_id": 2015, "contact_type": "corporation", "standing": 10.0},
]


def load_eve_entities():
    auth_to_eve_entities()
    map_to_category = {
        "alliance": EveEntity.CATEGORY_ALLIANCE,
        "corporation": EveEntity.CATEGORY_CORPORATION,
        "character": EveEntity.CATEGORY_CHARACTER,
    }
    for info in ALLIANCE_CONTACTS:
        EveEntity.objects.get_or_create(
            id=info["contact_id"],
            defaults={
                "category": map_to_category[info["contact_type"]],
                "name": f"dummy_{info['contact_id']}",
            },
        )


class LoadTestDataMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.character_1 = EveCharacter.objects.create(
            character_id=1001,
            character_name="Bruce Wayne",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        cls.corporation_1 = EveCorporationInfo.objects.create(
            corporation_id=cls.character_1.corporation_id,
            corporation_name=cls.character_1.corporation_name,
            member_count=99,
        )
        cls.alliance_1 = EveAllianceInfo.objects.create(
            alliance_id=cls.character_1.alliance_id,
            alliance_name=cls.character_1.alliance_name,
            executor_corp_id=cls.corporation_1.corporation_id,
        )
        cls.character_2 = EveCharacter.objects.create(
            character_id=1002,
            character_name="Clark Kent",
            corporation_id=2001,
            corporation_name="Wayne Technologies",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        cls.character_3 = EveCharacter.objects.create(
            character_id=1003,
            character_name="Lex Luthor",
            corporation_id=2003,
            corporation_name="Lex Corp",
            alliance_id=3003,
            alliance_name="Lex Holding",
        )
        cls.corporation_3 = EveCorporationInfo.objects.create(
            corporation_id=cls.character_3.corporation_id,
            corporation_name=cls.character_3.corporation_name,
            member_count=666,
        )
        cls.alliance_3 = EveAllianceInfo.objects.create(
            alliance_id=cls.character_3.alliance_id,
            alliance_name=cls.character_3.alliance_name,
            executor_corp_id=cls.corporation_3.corporation_id,
        )
        cls.character_4 = EveCharacter.objects.create(
            character_id=1004,
            character_name="Kara Danvers",
            corporation_id=2004,
            corporation_name="CatCo",
        )
        cls.corporation_4 = EveCorporationInfo.objects.create(
            corporation_id=cls.character_4.corporation_id,
            corporation_name=cls.character_4.corporation_name,
            member_count=1234,
        )
        cls.character_5 = EveCharacter.objects.create(
            character_id=1005,
            character_name="Peter Parker",
            corporation_id=2005,
            corporation_name="Daily Bugle",
        )
        cls.corporation_5 = EveCorporationInfo.objects.create(
            corporation_id=cls.character_5.corporation_id,
            corporation_name=cls.character_5.corporation_name,
            member_count=1234,
        )
        cls.character_6 = EveCharacter.objects.create(
            character_id=1099,
            character_name="Joe Doe",
            corporation_id=2005,
            corporation_name="Daily Bugle",
        )
        load_eve_entities()


def auth_to_eve_entities():
    """Creates EveEntity objects from existing Auth objects."""
    for obj in EveAllianceInfo.objects.all():
        EveEntity.objects.get_or_create(
            id=obj.alliance_id,
            defaults={
                "name": obj.alliance_name,
                "category": EveEntity.CATEGORY_ALLIANCE,
            },
        )
    for obj in EveCorporationInfo.objects.all():
        EveEntity.objects.get_or_create(
            id=obj.corporation_id,
            defaults={
                "name": obj.corporation_name,
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )
    for obj in EveCharacter.objects.all():
        EveEntity.objects.get_or_create(
            id=obj.character_id,
            defaults={
                "name": obj.character_name,
                "category": EveEntity.CATEGORY_CHARACTER,
            },
        )


@dataclass(frozen=True)
class EsiContactLabel:
    id: int
    name: str

    def __post_init__(self):
        object.__setattr__(self, "id", int(self.id))
        object.__setattr__(self, "name", str(self.name))

    def to_dict(self) -> dict:
        return {self.id: self.name}


@dataclass(unsafe_hash=True)
class EsiContact:
    """A contact in the ESI character contacts stub."""

    class ContactType(str, Enum):
        CHARACTER = "character"
        CORPORATION = "corporation"
        ALLIANCE = "alliance"

    contact_id: int
    contact_type: str
    standing: float
    label_ids: FrozenSet[int] = field(default_factory=frozenset)

    def __setattr__(self, prop, val):
        if prop == "contact_id":
            val = int(val)
        if prop == "standing":
            val = float(val)
        if prop == "label_ids":
            if not val:
                val = []
            val = self._clean_label_ids(val)
        if prop == "contact_type":
            if val not in self.ContactType:
                raise ValueError(f"Invalid contact_type: {val}")
        super().__setattr__(prop, val)

    def _clean_label_ids(self, label_ids):
        return frozenset([int(obj) for obj in label_ids])

    def to_esi_dict(self) -> dict:
        return {
            "contact_id": self.contact_id,
            "contact_type": self.ContactType(self.contact_type),
            "standing": self.standing,
            "label_ids": list(self.label_ids),
        }

    @classmethod
    def from_eve_entity(
        cls, eve_entity: EveEntity, standing: float, label_ids=None
    ) -> "EsiContact":
        """Create new instance from an EveEntity object."""
        contact_type_map = {
            EveEntity.CATEGORY_ALLIANCE: cls.ContactType.ALLIANCE,
            EveEntity.CATEGORY_CHARACTER: cls.ContactType.CHARACTER,
            EveEntity.CATEGORY_CORPORATION: cls.ContactType.CORPORATION,
        }
        return cls(
            contact_id=eve_entity.id,
            contact_type=contact_type_map[eve_entity.category],
            standing=standing,
            label_ids=label_ids,
        )


class EsiCharacterContactsStub:
    """Simulates the contacts for a character on ESI"""

    def __init__(self) -> None:
        self._contacts = dict()
        self._labels = dict()

    def setup_contacts(self, character_id: int, contacts: List[EsiContact]):
        self._contacts[character_id] = dict()
        if character_id not in self._labels:
            self._labels[character_id] = dict()
        for contact in contacts:
            if contact.label_ids:
                for label_id in contact.label_ids:
                    if label_id not in self.labels(character_id).keys():
                        raise ValueError(f"Invalid label_id: {label_id}")
            self._contacts[character_id][contact.contact_id] = copy.deepcopy(contact)

    def setup_labels(self, character_id: int, labels: List[EsiContact]):
        self._labels[character_id] = {obj.id: obj.name for obj in labels}

    def contacts(self, character_id: int) -> dict:
        return self._contacts[character_id].values()

    def character_contact(self, character_id: int, contact_id: int) -> EsiContact:
        return self._contacts[character_id][contact_id]

    def labels(self, character_id: int) -> dict:
        return self._labels[character_id] if character_id in self._labels else dict()

    def setup_esi_mock(self, mock_esi):
        """Sets the mock for ESI to this object."""
        mock_esi.client.Contacts.get_characters_character_id_contacts.side_effect = (
            self._esi_get_characters_character_id_contacts
        )
        mock_esi.client.Contacts.delete_characters_character_id_contacts.side_effect = (
            self._esi_delete_characters_character_id_contacts
        )
        mock_esi.client.Contacts.post_characters_character_id_contacts = (
            self._esi_post_characters_character_id_contacts
        )
        mock_esi.client.Contacts.put_characters_character_id_contacts = (
            self._esi_put_characters_character_id_contacts
        )
        mock_esi.client.Contacts.get_characters_character_id_contacts_labels = (
            self._esi_get_characters_character_id_contacts_labels
        )

    def _esi_get_characters_character_id_contacts(self, character_id, token, page=None):
        if character_id in self._contacts:
            contacts = [
                obj.to_esi_dict() for obj in self._contacts[character_id].values()
            ]
        else:
            contacts = []
        return BravadoOperationStub(contacts)

    def _esi_get_characters_character_id_contacts_labels(
        self, character_id, token, page=None
    ):
        if character_id in self._labels:
            labels = [
                {"label_id": k, "label_name": v}
                for k, v in self._labels[character_id].items()
            ]
        else:
            labels = []
        return BravadoOperationStub(labels)

    def _esi_post_characters_character_id_contacts(
        self, character_id, contact_ids, standing, token, label_ids=None
    ):
        self._check_label_ids_valid(character_id, label_ids)
        contact_type_map = {
            EveEntity.CATEGORY_CHARACTER: EsiContact.ContactType.CHARACTER,
            EveEntity.CATEGORY_CORPORATION: EsiContact.ContactType.CORPORATION,
            EveEntity.CATEGORY_ALLIANCE: EsiContact.ContactType.ALLIANCE,
        }
        if character_id not in self._contacts:
            self._contacts[character_id] = dict()
        for contact_id in contact_ids:
            eve_entity = EveEntity.objects.get(id=contact_id)
            self._contacts[character_id][contact_id] = EsiContact(
                contact_id=contact_id,
                contact_type=contact_type_map[eve_entity.category],
                standing=standing,
                label_ids=label_ids,
            )
        return BravadoOperationStub([])

    def _esi_put_characters_character_id_contacts(
        self, character_id, contact_ids, standing, token, label_ids=None
    ):
        self._check_label_ids_valid(character_id, label_ids)
        for contact_id in contact_ids:
            self._contacts[character_id][contact_id].standing = standing
            if label_ids:
                if not self._contacts[character_id][contact_id].label_ids:
                    self._contacts[character_id][contact_id].label_ids = label_ids
                else:
                    self._contacts[character_id][contact_id].label_ids += label_ids
        return BravadoOperationStub([])

    def _esi_delete_characters_character_id_contacts(
        self, character_id, contact_ids, token
    ):
        for contact_id in contact_ids:
            del self._contacts[character_id][contact_id]
        return BravadoOperationStub([])

    def _check_label_ids_valid(self, character_id, label_ids):
        if label_ids:
            for label_id in label_ids:
                if label_id not in self.labels(character_id).keys():
                    raise ValueError(f"Invalid label_id: {label_id}")
